/*
   +----------------------------------------------------------------------+
   | HipHop for PHP                                                       |
   +----------------------------------------------------------------------+
   | Copyright (c) 2010- Facebook, Inc. (http://www.facebook.com)         |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_01.txt                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
*/

#include <runtime/eval/ast/instanceof_expression.h>
#include <runtime/eval/ast/name.h>

namespace HPHP {
namespace Eval {
///////////////////////////////////////////////////////////////////////////////

InstanceOfExpression::InstanceOfExpression(EXPRESSION_ARGS, ExpressionPtr obj,
                                           NamePtr name)
  : Expression(KindOfInstanceOfExpression, EXPRESSION_PASS),
   m_obj(obj), m_name(name) {}

Expression *InstanceOfExpression::optimize(VariableEnvironment &env) {
  Eval::optimize(env, m_obj);
  Eval::optimize(env, m_name);
  return NULL;
}

Variant InstanceOfExpression::eval(VariableEnvironment &env) const {
  Variant obj(m_obj->eval(env));
  String sname(m_name->get());
  if (sname == "self") {
    return instanceOf(obj, FrameInjection::GetClassName(false));
  } else if (sname == "parent") {
    return instanceOf(obj, FrameInjection::GetParentClassName(false));
  }
  String name(m_name->get(env));
  return instanceOf(obj, name);
}

void InstanceOfExpression::dump(std::ostream &out) const {
  m_obj->dump(out);
  out << " instanceof ";
  m_name->dump(out);
}

///////////////////////////////////////////////////////////////////////////////
}
}

